from logging import DEBUG, Formatter, Logger, getLogger
from logging.handlers import RotatingFileHandler

from gardnr import settings


def enable_debugging() -> None:
    settings.LOG_LEVEL = DEBUG


def get_logger() -> Logger:
    return getLogger('gardnr')


def setup() -> None:
    file_rotation = RotatingFileHandler(settings.LOG_FILE,
                                        maxBytes=settings.LOG_FILE_SIZE,
                                        backupCount=settings.LOG_FILE_COUNT)
    formatter = Formatter(fmt='%(asctime)s - %(levelname)s - '
                          '%(module)s - %(message)s')

    file_rotation.setFormatter(formatter)

    logger = get_logger()
    logger.setLevel(settings.LOG_LEVEL)
    logger.addHandler(file_rotation)


def debug(msg: str) -> None:
    get_logger().debug(msg)


def info(msg: str) -> None:
    get_logger().info(msg)


def warning(msg: str) -> None:
    get_logger().warning(msg)


def error(msg: str) -> None:
    get_logger().error(msg)


def critical(msg: str) -> None:
    get_logger().critical(msg)


def exception(msg: str) -> None:
    get_logger().exception(msg)



#!/usr/bin/env python
"""
Runs the scheduler and workers of scheduled tasks for drivers
"""
from datetime import timedelta
from typing import List, TextIO, Tuple

from apscheduler.schedulers.blocking import BlockingScheduler

from gardnr import constants, drivers, grow, logger, models, reflection, tasks, settings


trigger_bounds = []  # List[TriggerBound]
active_trigger_bounds = []  # List[TriggerBound]


def main() -> None:

    logger.setup()

    logger.info('Starting automata')

    models.initialize_db()

    reflection.add_driver_path(settings.DRIVER_PATH)

    scheduler = BlockingScheduler()

    schedules = models.Schedule.select()\
        .where(models.Schedule.disabled == False)
    for schedule in schedules:

        driver_names = _build_driver_dict(schedule)

        scheduler.add_job(
            driver_worker,
            'cron',
            [*driver_names],
            minute=schedule.minute,
            hour=schedule.hour,
            day=schedule.day_of_month,
            month=schedule.month,
            day_of_week=schedule.day_of_week,
        )

    tracked_grow = grow.get_tracked_grow()

    if tracked_grow:
        with open(settings.GROW_RECIPE) as recipe:
            _build_trigger_bounds(tracked_grow, recipe)

        scheduler.add_job(
            bound_checker,
            'interval',
            minutes=settings.BOUND_CHECK_FREQUENCY
        )

    scheduler.start()


def _build_driver_dict(
        schedule: models.Schedule
) -> Tuple[List[str], List[str], List[str], List[str]]:
    """
    Helper function to build a dict with the drivers names categorized
    into what type and action they require
    """

    power_on_names = []
    power_off_names = []
    sensor_names = []
    exporter_names = []

    for driver_schedule in schedule.driver_schedules:
        driver = driver_schedule.driver

        if driver.type == constants.POWER:
            if driver_schedule.power_on:
                power_on_names.append(driver.name)
            else:
                power_off_names.append(driver.name)
        elif driver.type == constants.SENSOR:
            sensor_names.append(driver.name)
        elif driver.type == constants.EXPORTER:
            exporter_names.append(driver.name)

    return power_on_names, power_off_names, sensor_names, exporter_names


def driver_worker(
        power_on_names: List[str],
        power_off_names: List[str],
        sensor_names: List[str],
        exporter_names: List[str]
) -> None:

    if power_on_names:
        power_on_drivers = reflection.load_active_drivers(
            constants.POWER,
            power_on_names
        )
        tasks.power_on(power_on_drivers)

    if power_off_names:
        power_off_drivers = reflection.load_active_drivers(
            constants.POWER,
            power_off_names
        )
        tasks.power_off(power_off_drivers)

    if sensor_names:
        sensors = reflection.load_active_drivers(
            constants.SENSOR,
            sensor_names
        )
        tasks.read(sensors)

    if exporter_names:
        exporters = reflection.load_active_drivers(
            constants.EXPORTER,
            exporter_names
        )
        tasks.write(exporters)


class TriggerBound:
    """
    A class to dynamically link configured triggers with a grow recipe bound
    """
    def __init__(
            self,
            trigger: models.Trigger,
            bound: float,
            power_driver: drivers.Power
    ) -> None:
        self.trigger = trigger
        self.bound = bound
        self.power_driver = power_driver


def _build_trigger_bounds(
        tracked_grow: models.Grow,
        recipe: TextIO
) -> None:

    active_triggers = models.Trigger.select().where(
        models.Trigger.disabled == False)

    for trigger in active_triggers:
        metric_bound = grow.get_metric_bound(
            recipe, tracked_grow, trigger.metric.topic, trigger.metric.type)

        if metric_bound:
            power_driver = reflection.load_driver(trigger.power_driver)

            if metric_bound.min is not None and not trigger.upper_bound:
                trigger_bounds.append(TriggerBound(
                    trigger, metric_bound.min, power_driver))
            elif metric_bound.max is not None and trigger.upper_bound:
                trigger_bounds.append(TriggerBound(
                    trigger, metric_bound.max, power_driver))


def bound_checker() -> None:

    def _is_out_of_bound(tb: TriggerBound, log: models.MetricLog) -> bool:

        if tb.trigger.upper_bound and\
           log.value > tb.bound:
            return True

        elif not tb.trigger.upper_bound and\
                log.value < tb.bound:
            return True

        return False

    # check to reverse the trigger action
    for tb in active_trigger_bounds:
        # check since the last run
        latest_log = tb.trigger.metric.get_latest_log(
            timedelta(minutes=settings.BOUND_CHECK_FREQUENCY))

        if latest_log and not _is_out_of_bound(tb, latest_log):
            log_str = ('Turning {{on_off}} {power_driver_name} to reverse '
                       'the action for {metric_name} being too '
                       '{low_high}').format(
                           power_driver_name=tb.trigger.power_driver.name,
                           metric_name=tb.trigger.metric.name,
                           low_high=('high' if tb.trigger.upper_bound
                                     else 'low'))

            reflection.load_driver(tb.trigger.power_driver)

            if tb.trigger.power_on:
                logger.info(log_str.format(on_off='off'))
                tb.power_driver.off()
            else:
                logger.info(log_str.format(on_off='on'))
                tb.power_driver.on()

            active_trigger_bounds.remove(tb)

    for tb in trigger_bounds:
        latest_log = tb.trigger.metric.get_latest_log(
            timedelta(minutes=settings.BOUND_CHECK_FREQUENCY))

        if latest_log and _is_out_of_bound(tb, latest_log):
            log_str = ('Turning {{on_off}} {power_driver_name} because '
                       '{metric_name} is too {low_high}').format(
                           power_driver_name=tb.trigger.power_driver.name,
                           metric_name=tb.trigger.metric.name,
                           low_high=('high' if tb.trigger.upper_bound
                                     else 'low'))

            if tb.trigger.power_on:
                logger.info(log_str.format(on_off='on'))
                tb.power_driver.on()
            else:
                logger.info(log_str.format(on_off='off'))
                tb.power_driver.off()

            active_trigger_bounds.append(tb)


if __name__ == '__main__':
    main()



import logging
import os

from gardnr import constants

# when TEST_MODE is enabled, data is stored in a temporary in-memory database
TEST_MODE = False

LOCAL_DB = 'gardnr.db'
DRIVER_PATH = os.getcwd()

# logging
LOG_FILE = 'gardnr.log'
LOG_LEVEL = logging.INFO
LOG_FILE_SIZE = 10485760  # 10MB
LOG_FILE_COUNT = 1

UPLOAD_PATH = 'uploaded'

# l10n
TEMPERATURE_UNIT = constants.CELSIUS

GROW_RECIPE = None
BOUND_CHECK_FREQUENCY = 5  # in minutes

# can be overwritten incase there are more verbose templates
TEMPLATE_DIRECTORY = 'templates'
SENSOR_DRIVER_TEMPLATE = 'sensor_driver.py'
EXPORTER_DRIVER_TEMPLATE = 'exporter_driver.py'
POWER_DRIVER_TEMPLATE = 'power_driver.py'

try:
    # flake8: noqa
    # pylint: disable=wildcard-import
    from settings_local import *  # type: ignore
except ImportError:
    pass



from typing import List

import peewee

from gardnr import constants, drivers, logger, models


def _exported_logs(driver_model: models.Driver) -> peewee.ModelSelect:
    """Helper function to get the export logs associated with a driver"""

    return models.ExportLog.select(models.ExportLog.metric_log.id)\
        .join(models.MetricLog)\
        .where(models.ExportLog.driver == driver_model)


def get_all_logs(driver_model: models.Driver) -> peewee.ModelSelect:
    """All logs that have not been exported with driver_model"""

    return models.MetricLog.select().where(
        models.MetricLog.id.not_in(_exported_logs(driver_model))
    )


def get_whitelisted_logs(
        driver_model: models.Driver,
        whitelist: List[str]
) -> peewee.ModelSelect:
    """
    Logs from whitelisted metrics that have not been exported with
    driver_model
    """

    return models.MetricLog.select()\
        .join(models.Metric)\
        .where(models.MetricLog.id.not_in(_exported_logs(driver_model))
               & models.MetricLog.metric.type.in_(whitelist))


def get_not_blacklisted_logs(
        driver_model: models.Driver,
        blacklist: List[str]
) -> peewee.ModelSelect:
    """
    Logs from whitelisted metrics that have not been exported with
    driver_model
    """

    return models.MetricLog.select()\
        .join(models.Metric)\
        .where(models.MetricLog.id.not_in(_exported_logs(driver_model))
               & models.MetricLog.metric.type.not_in(blacklist))


def write(exporters: List[drivers.Exporter]) -> None:
    """Upload logs in local DB to web server."""

    for exporter in exporters:

        if exporter.whitelist:
            logs = get_whitelisted_logs(exporter.model, exporter.whitelist)
        elif exporter.blacklist:
            logs = get_not_blacklisted_logs(exporter.model, exporter.blacklist)
        else:
            logs = get_all_logs(exporter.model)

        # no new logs to export
        if not logs:
            continue

        # store the failed logs during export
        failed_log_ids = []  # type: List[UUID]

        try:
            exporter.export([log for log in logs])  # type: ignore
        except Exception as e:  # pylint: disable=broad-except
            logger.exception('Error transmitting')

            # If the exporter sets the failed_logs field in the exception
            # use it, otherwise assume all logs failed
            failed_logs = getattr(e, 'failed_logs', [])

            # If there are no failed logs specified, skip export logging
            if not failed_logs:
                continue

            failed_log_ids = [log.id for log in failed_logs]

        for log in logs:
            if log.id in failed_log_ids:
                continue

            models.ExportLog.create(metric_log=log, driver=exporter.model)



from grow_recipe.constants import *  # noqa F403

NOTES = 'notes'
IMAGE = 'image'
POWER = 'power'
GARDNR_METRICS = {'notes', 'image', 'power'}
metrics = metrics.union(GARDNR_METRICS)

MANUAL_OVERRIDE_CONFIG = 'manual'


SENSOR = 'sensor'
POWER = 'power'
EXPORTER = 'exporter'
drivers = {SENSOR, POWER, EXPORTER}


CELSIUS = 'c'
FAHRENHEIT = 'f'
t9e_untis = {CELSIUS, FAHRENHEIT}



from typing import List

from gardnr import drivers, logger, metrics


def power_on(power_devices: List[drivers.Power]) -> None:
    for power_device in power_devices:
        logger.info('Powered on {}'.format(power_device.model.name))
        power_device.on()


def power_off(power_devices: List[drivers.Power]) -> None:
    for power_device in power_devices:
        logger.info('Powering off {}'.format(power_device.model.name))
        power_device.off()



import importlib
import sys
from typing import List, Optional

from gardnr import constants, drivers, models


def add_driver_path(path: str) -> None:
    """Adds path to PYTHONPATH"""
    sys.path.append(path)


def get_fully_qualname(cls: type) -> str:
    """module with qualified name"""
    return '{module}:{qualname}'.format(module=cls.__module__,
                                        qualname=cls.__qualname__)


def get_class_type(fully_qualname: str) -> type:
    """dynamically loads a class type"""
    module_name, class_name = fully_qualname.split(':')
    class_type = getattr(importlib.import_module(module_name),
                         class_name)

    if not class_type:
        raise NameError('Unable to load classfrom fully qualified '
                        'name {name}'.format(name=fully_qualname))

    return class_type


def load_driver(driver_model: models.Driver) -> drivers.Driver:
    """factory for configured driver objects"""

    class_type = get_class_type(driver_model.fully_qualname)

    return class_type(driver_model)


def load_active_drivers(
        driver_type: Optional[str] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None
) -> List[drivers.Driver]:
    """Loads a list of Drivers by name that are not disabled"""

    driver_models = models.Driver.select().where(
        models.Driver.disabled == False
    )

    if driver_type:
        driver_models = driver_models.where(
            models.Driver.type == driver_type
        )

    if include:
        driver_models = driver_models.where(models.Driver.name.in_(include))
    elif exclude:
        driver_models = driver_models.where(models.Driver.name.not_in(exclude))

    return [load_driver(driver_model) for driver_model in driver_models]


def get_base_type(driver_class: type) -> str:
    if issubclass(driver_class, drivers.Sensor):
        return constants.SENSOR
    if issubclass(driver_class, drivers.Power):
        return constants.POWER
    elif issubclass(driver_class, drivers.Exporter):
        return constants.EXPORTER

    raise ValueError('Driver base type not found')



import os
from typing import Any
from uuid import uuid4

from gardnr import constants, models, settings


def create_metric_log(metric_name: str, value: Any) -> models.MetricLog:
    """sets up common metric log fields"""

    metric = models.Metric.get(models.Metric.name == metric_name)

    return models.MetricLog.create(id=uuid4(), metric=metric, value=value)


def create_file_log(metric_name: str,
                    blob: bytes,
                    extension: str) -> models.MetricLog:
    """
    Creates a special metric log for blob (bytes) type
    metrics, i.e. images and videos. The value of the metric
    is the path of the stored file. The blob is then written
    to disk.

    NOTE: extension should be in the format '.jpg' which includes
    the beginning .
    """

    metric = models.Metric.get(models.Metric.name == metric_name)

    uuid = uuid4()

    uploaded_file_name = '{uuid}{extension}'.format(
        uuid=str(uuid).replace('-', ''),
        extension=extension
    )

    full_path = os.path.join(settings.UPLOAD_PATH, uploaded_file_name)

    open(full_path, 'wb').write(blob)

    return models.MetricLog.create(id=uuid,
                                   metric=metric,
                                   value=full_path)


class MetricBase:

    @staticmethod
    def standardize(value: float) -> Any:
        """
        Check settings to convert localized metric value to standardized one
        """
        return value

    @staticmethod
    def validate(value: Any) -> bool:
        del value
        return True


class TemperatureMetric(MetricBase):

    @staticmethod
    def fahrenheit_to_celsius(fahrenheit_temperature: float) -> float:
        return (fahrenheit_temperature - 32) * 5/9

    @staticmethod
    def standardize(value: float) -> float:
        """Creates temperature log"""

        if settings.TEMPERATURE_UNIT == constants.FAHRENHEIT:
            value = TemperatureMetric.fahrenheit_to_celsius(value)

        return value


class HumidityMetric(MetricBase):

    @staticmethod
    def validate(value: float) -> bool:
        return value >= 0 and value <= 1


def standardize_metric(metric_type: str, value: float) -> float:
    """Convert metric value from localized version to standardized one"""

    if metric_type == constants.T9E:
        value = TemperatureMetric.standardize(value)

    return value



from typing import List
from uuid import uuid4

import pytest

from gardnr import constants, drivers, models, tasks
from tests import utils


@pytest.mark.usefixtures('test_env')
def test_export_basic():
    """Test export a single log"""

    sensor = utils.create_and_load_air_temperature_sensor()

    tasks.read([sensor])

    exporter = utils.create_and_load_exporter()

    export_logs1 = models.ExportLog.select()
    assert export_logs1.count() == 0
    assert utils.MockExporter.call_count == 0

    tasks.write([exporter])

    export_logs2 = models.ExportLog.select()
    assert export_logs2.count() == 1
    assert utils.MockExporter.call_count == 1


@pytest.mark.usefixtures('test_env')
def test_export_missing_records():
    """
    You need to join to the relationship model whose attributes are being
    selected
    """

    sensor = utils.create_and_load_air_temperature_sensor()
    exporter = utils.create_and_load_exporter()

    tasks.read([sensor])
    tasks.write([exporter])

    export_logs1 = models.ExportLog.select()
    assert export_logs1.count() == 1
    assert utils.MockExporter.call_count == 1

    tasks.read([sensor])
    tasks.write([exporter])

    export_logs2 = models.ExportLog.select()
    assert export_logs2.count() == 2
    assert utils.MockExporter.call_count == 2


@pytest.mark.usefixtures('test_env')
def test_export_new_exporter():

    sensor = utils.create_and_load_air_temperature_sensor()

    tasks.read([sensor])

    exporter = utils.create_and_load_exporter()

    export_logs1 = models.ExportLog.select()
    assert export_logs1.count() == 0
    assert utils.MockExporter.call_count == 0

    tasks.write([exporter])

    export_logs2 = models.ExportLog.select()
    assert export_logs2.count() == 1
    assert utils.MockExporter.call_count == 1

    exporter2 = utils.create_and_load_exporter('new-exporter')

    tasks.write([exporter, exporter2])

    export_logs3 = models.ExportLog.select()
    assert export_logs3.count() == 2
    assert utils.MockExporter.call_count == 2


class FailingExporterSpecify(drivers.Exporter):
    def export(self, logs: List[models.MetricLog]) -> None:
        e = Exception()
        e.failed_logs = [logs[0]]
        raise e


@pytest.mark.usefixtures('test_env')
def test_export_failed_specify():

    sensor = utils.create_and_load_air_temperature_sensor()

    tasks.read([sensor])
    tasks.read([sensor])

    exporter = utils.create_and_load_exporter(
        driver_type=FailingExporterSpecify)

    tasks.write([exporter])

    export_logs = models.ExportLog.select()
    assert export_logs.count() == 1


class FailingExporterAll(drivers.Exporter):
    def export(self, logs: List[models.MetricLog]) -> None:
        raise Exception()


@pytest.mark.usefixtures('test_env')
def test_export_failed_all():

    sensor = utils.create_and_load_air_temperature_sensor()

    tasks.read([sensor])

    exporter = utils.create_and_load_exporter(driver_type=FailingExporterAll)

    tasks.write([exporter])

    export_logs = models.ExportLog.select()
    assert export_logs.count() == 0


class WhitelistExporter(utils.MockExporter):
    whitelist = [constants.HUMIDITY]


@pytest.mark.usefixtures('test_env')
def test_whitelisted_write():

    temp_metric = utils.create_air_temperature_metric()
    models.MetricLog.create(id=uuid4(), metric=temp_metric, value=0)

    humid_metric = models.Metric.create(id=uuid4(),
                                        name='test-humidity',
                                        topic=constants.AIR,
                                        type=constants.HUMIDITY)
    whitelisted_log = models.MetricLog.create(id=uuid4(),
                                              metric=humid_metric,
                                              value=0)

    whitelist_exporter = utils.create_and_load_exporter('whitelist-exporter',
                                                        WhitelistExporter)

    tasks.write([whitelist_exporter])

    export_logs = models.ExportLog.select()
    assert export_logs.count() == 1
    assert export_logs[0].metric_log.id == whitelisted_log.id


class BlacklistExporter(utils.MockExporter):
    blacklist = [constants.HUMIDITY]


@pytest.mark.usefixtures('test_env')
def test_blacklisted_write():

    temp_metric = utils.create_air_temperature_metric()
    models.MetricLog.create(id=uuid4(), metric=temp_metric, value=0)

    humid_metric = models.Metric.create(id=uuid4(),
                                        name='test-humidity',
                                        topic=constants.AIR,
                                        type=constants.HUMIDITY)
    blacklisted_log = models.MetricLog.create(id=uuid4(),
                                              metric=humid_metric,
                                              value=0)

    blacklist_exporter = utils.create_and_load_exporter('blacklist-exporter',
                                                        BlacklistExporter)

    tasks.write([blacklist_exporter])

    export_logs = models.ExportLog.select()
    assert export_logs.count() == 1
    assert export_logs[0].metric_log.id != blacklisted_log.id



import pytest

from gardnr import models, tasks
from tests import utils


@pytest.mark.usefixtures('test_env')
def test_power_on():

    power_device = utils.create_and_load_power_device()

    assert power_device.on_count == 0

    tasks.power_on([power_device])

    assert power_device.on_count == 1


@pytest.mark.usefixtures('test_env')
def test_power_off():

    power_device = utils.create_and_load_power_device()

    assert power_device.off_count == 0

    tasks.power_off([power_device])

    assert power_device.off_count == 1



import pytest

from gardnr import models, tasks
from tests import utils


@pytest.mark.usefixtures('test_env')
def test_read_sensor():
    """basic test for reading a mock sensor."""

    sensor = utils.create_and_load_air_temperature_sensor()

    tasks.read([sensor])

    logs = models.MetricLog.select()

    assert logs.count() == 1
    assert logs[0].metric.name == utils.TEST_METRIC
    assert logs[0].value == utils.TEST_TEMPERATURE



from datetime import datetime
from typing import Optional, TextIO
from uuid import uuid4

import grow_recipe

from gardnr import models, settings


class GrowAlreadyActiveError(Exception):
    pass


class NoActiveGrowError(Exception):
    pass


def get_active() -> Optional[models.Grow]:
    """
    Returns a grow that is currently in progress,
    A grow in progress is one with an 'end' set to NULL
    """
    # pylint: disable=singleton-comparison
    return models.Grow.get_or_none(models.Grow.end == None)  # noqa: E711


def get_tracked_grow() -> Optional[models.Grow]:
    """
    Only returns the active grow if there is recipe set
    """
    if settings.GROW_RECIPE:
        return get_active()

    return None


def get_current_stage(active_grow: models.Grow) -> Optional[str]:

    if settings.GROW_RECIPE:
        with open(settings.GROW_RECIPE) as recipe:
            return grow_recipe.get_grow_stage(recipe, active_grow.start)

    return None


def get_metric_bound(
        recipe: TextIO,
        tracked_grow: models.Grow,
        metric_topic: str,
        metric_type: str
) -> Optional[grow_recipe.query.find_metric_value.Metric]:
    return grow_recipe.get_metric(
        recipe, metric_topic, metric_type, tracked_grow.start)


def start() -> None:
    if get_active():
        raise GrowAlreadyActiveError()

    models.Grow.create(id=uuid4())


def end() -> None:
    active_grow = get_active()

    if not active_grow:
        raise NoActiveGrowError()

    active_grow.end = datetime.utcnow()
    active_grow.save()



from gardnr import drivers, logger, metrics


class Sensor(drivers.Sensor):
    """
    Add code to interface with physical or virtual sensors here.
    """

    def setup(self):
        """
        Add configuration here
        """

        # remove the next line and add code
        pass

    def read(self):
        """
        Example log:

        metrics.create_metric_log('my-metric', 42)


        Example image log:

        metrics.create_image_log(
            self.metric_name,
            image_bytes,
            extension=self.image_file_extension
        )
        """

        # remove the next line and add code
        pass



"""Driver and log models."""
import json
import uuid
from datetime import datetime, timedelta
from typing import Dict, Optional, List

import cron_descriptor
from peewee import (BlobField, BooleanField, DateTimeField, FloatField,
                    ForeignKeyField, Model, ModelSelect, SqliteDatabase,
                    TextField, UUIDField)

from gardnr import constants, settings

# initialize with None so run-time variables can be used
# set the connection string, reference: https://goo.gl/ms2oHP
_db = SqliteDatabase(None)


def initialize_db() -> None:
    """
    Creates any missing tables in the database
    """

    if settings.TEST_MODE:
        _db.init(':memory:')
    else:
        _db.init(settings.LOCAL_DB)

    # incase there is a connection issue it will be found here
    # instead of downstream
    _db.connect()

    _db.create_tables(BaseModel.__subclasses__(), safe=True)


class BaseModel(Model):
    class Meta:
        database = _db


class JSONField(TextField):

    def db_value(self, value: Dict) -> str:
        return json.dumps(value)

    def python_value(self, value: Optional[str]) -> Optional[Dict]:
        if value is not None:
            return json.loads(value)
        return None


class Metric(BaseModel):

    id = UUIDField(primary_key=True)
    name = TextField(unique=True)

    topic = TextField(choices=[(topic, topic)
                               for topic in constants.topics])
    type = TextField(choices=[(metric, metric)
                              for metric in constants.metrics])

    manual = BooleanField(default=False)

    disabled = BooleanField(default=False)

    def get_latest_log(
            self,
            cutoff: Optional[timedelta] = None
    ) -> Optional['MetricLog']:
        """
        Returns the latest log for a given log. Optionally, provide a
        cutoff date for the latest the log can be from.
        """

        latest_log = MetricLog.select()\
            .join(Metric)\
            .where(Metric.name == self.name)\
            .order_by(MetricLog.timestamp.desc())\
            .limit(1)

        if cutoff:
            latest_log = latest_log.where(
                MetricLog.timestamp >= datetime.utcnow() - cutoff)

        if not latest_log:
            return None

        return latest_log[0]


class Driver(BaseModel):
    """
    Information about a device.

    Config holds information on how to access data from the device.
    It can be customized
    """

    name = TextField(unique=True)
    type = TextField(choices=[(driver, driver)
                              for driver in constants.drivers])
    fully_qualname = TextField()

    config = JSONField(null=True)

    disabled = BooleanField(default=False)


class Schedule(BaseModel):
    """
    cron-style schedule to execute drivers on
    """

    name = TextField(unique=True)

    disabled = BooleanField(default=False)

    minute = TextField()
    hour = TextField()
    day_of_month = TextField()
    month = TextField()
    day_of_week = TextField()

    @property
    def crontab(self):
        """
        Converts model fields to crontab syntax
        """
        return ('{minute} {hour} {day_of_month} '
                '{month} {day_of_week}').format(
                    minute=self.minute,
                    hour=self.hour,
                    day_of_month=self.day_of_month,
                    month=self.month,
                    day_of_week=self.day_of_week)

    def transcribe(self) -> str:
        """
        converts crontab syntax to english
        """
        return cron_descriptor.get_description(self.crontab)


class DriverSchedule(BaseModel):
    """
    Many-to-many relationship between drivers and schedules
    peewee docs: https://goo.gl/8SB4iY
    """

    driver = ForeignKeyField(Driver)
    schedule = ForeignKeyField(Schedule, backref='driver_schedules')

    # used for power devices to be either turned on or off
    power_on = BooleanField(null=True, default=None)


class MetricLog(BaseModel):
    id = UUIDField(primary_key=True)

    timestamp = DateTimeField(default=datetime.utcnow)

    longitude = FloatField(null=True)
    latitude = FloatField(null=True)
    elevation = FloatField(null=True)

    value = BlobField()

    metric = ForeignKeyField(Metric)


class ExportLog(BaseModel):
    """
    Many-to-many relationship which logs occurrences of metric logs being
    exported
    """

    metric_log = ForeignKeyField(MetricLog)
    driver = ForeignKeyField(Driver)


class Trigger(BaseModel):
    """
    A rule for what power device to either turn on or off a when a metric's
    value hits either an upper or lower bound (too_high)
    """

    metric = ForeignKeyField(Metric)
    # either the upper bound or lower bound for the metric
    upper_bound = BooleanField()

    power_driver = ForeignKeyField(Driver)
    # either turn on or off for the power driver
    power_on = BooleanField()

    disabled = BooleanField(default=False)


class Grow(BaseModel):
    id = UUIDField(primary_key=True)

    start = DateTimeField(default=datetime.utcnow)
    end = DateTimeField(null=True)



from gardnr import drivers, logger


class Power(drivers.Power):
    """
    Add code to interface with physical powered devices.
    """

    def setup(self):
        """
        Add configuration here
        """

        # remove the next line and add code
        pass

    def on(self):

        """
        Communicate with a powered device

        Example:

        import subprocess
        command = 'gpio mode 0 out; gpio write 0 1'
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        """

        # remove the next line and add code
        pass

    def off(self):
        pass



#!/usr/bin/env python
"""
Contains entry point to the program and command line parsing
"""

import argparse
import sys
import uuid
import os
import shutil
from datetime import datetime
from typing import List, Optional, Tuple

from gardnr import (constants, grow, logger, models,
                    reflection, settings, tasks)

LOGO = (r"""
 #####      #     ######   ######   #     #  ######
#     #    # #    #     #  #     #  ##    #  #     #
#         #   #   #     #  #     #  # #   #  #     #
#  ####  #     #  ######   #     #  #  #  #  ######
#     #  #######  #   #    #     #  #   # #  #   #
#     #  #     #  #    #   #     #  #    ##  #    #
 #####   #     #  #     #  ######   #     #  #     #
""")


def main() -> None:
    logger.setup()

    parser, args = create_and_run_parser(sys.argv[1:])

    # default print help
    if len(sys.argv) == 1:
        parser.print_help()
        return

    if args.test:
        settings.TEST_MODE = True
        logger.enable_debugging()

    # wait to see if in test mode to initialize the database
    models.initialize_db()

    reflection.add_driver_path(settings.DRIVER_PATH)

    # execute the command
    args.func(args)


def _bound_to_bool(bound: str) -> bool:
    if bound == 'min':
        return False

    if bound == 'max':
        return True

    raise ValueError


def _bool_to_bound(upper_bound: bool) -> str:
    return 'max' if upper_bound else 'min'


def _power_to_bool(power: str) -> bool:
    if power == 'on':
        return True

    if power == 'off':
        return False

    raise ValueError


def _bool_to_power(powered_on: bool) -> str:
    return '[on]' if powered_on else '[off]' if powered_on is not None else ''


BOUND_INVALID_HELP = ('Invalid value for a bound, either specify min or'
                      ' max')
POWER_INVALID_HELP = ('Invalid value for power, either specify on or'
                      ' off')
NO_POWER_HELP = 'Power devices need the power argument specified'


def _copy_template(template_file: str, dest: Optional[str]) -> None:
    """
    Copies a given template file into a given destination file.
    If no file is given, uses the template file name and places
    the file in the current working directory.
    """

    gardnr_directory = os.path.dirname(os.path.abspath(__file__))
    template = ('{gardnr_directory}/{template_directory}/'
                '{template_file}').format(
                    gardnr_directory=gardnr_directory,
                    template_directory=settings.TEMPLATE_DIRECTORY,
                    template_file=template_file
                )

    if not dest:
        dest = '.'

    shutil.copy(template, dest)


def new_power(args: argparse.Namespace) -> None:
    _copy_template(settings.POWER_DRIVER_TEMPLATE, args.filename)


def new_sensor(args: argparse.Namespace) -> None:
    _copy_template(settings.SENSOR_DRIVER_TEMPLATE, args.filename)


def new_exporter(args: argparse.Namespace) -> None:
    _copy_template(settings.EXPORTER_DRIVER_TEMPLATE, args.filename)


def add_metric(args: argparse.Namespace) -> None:
    metric = models.Metric(id=uuid.uuid4(),
                           name=args.name,
                           topic=args.topic,
                           type=args.type,
                           manual=args.manual)
    metric.save(force_insert=True)

    print('{topic} {type} metric {name} added'.format(topic=metric.topic,
                                                      type=metric.type,
                                                      name=metric.name))


def add_driver(args: argparse.Namespace) -> None:
    # identify the driver
    driver_class = reflection.get_class_type(args.classname)

    driver_base = reflection.get_base_type(driver_class)

    driver = models.Driver(type=driver_base,
                           name=args.name,
                           fully_qualname=args.classname)

    if args.config is not None:
        driver.config = args.config

    driver.save()

    print('{type} driver {name} added'.format(type=driver_class,
                                              name=driver.name))


def add_schedule(args: argparse.Namespace) -> None:

    schedule = models.Schedule(
        name=args.name,
        minute=args.minute,
        hour=args.hour,
        day_of_month=args.day_of_month,
        month=args.month,
        day_of_week=args.day_of_week)

    if not args.yes:
        print('You entered: {}'.format(schedule.transcribe()))
        print('Does this look correct? ([y]/n)')
        answer = input()
        if answer == 'n':
            return

    schedule.save()

    print('schedule {} added'.format(schedule.name))


def add_trigger(args: argparse.Namespace) -> None:

    metric = models.Metric.get(models.Metric.name == args.metric)

    try:
        upper_bound = _bound_to_bool(args.bound)
    except ValueError:
        print(BOUND_INVALID_HELP)
        return

    power_driver = models.Driver.get(
        (models.Driver.name == args.driver) &
        (models.Driver.type == constants.POWER)
    )

    try:
        powered_on = _power_to_bool(args.power)
    except ValueError:
        print(POWER_INVALID_HELP)
        return

    models.Trigger.create(metric=metric,
                          upper_bound=upper_bound,
                          power_driver=power_driver,
                          power_on=powered_on)

    print('Trigger added for turning {driver} {power} when {metric} '
          'hits {bound} '.format(
              power=args.power,
              driver=args.driver,
              bound=args.bound,
              metric=args.metric))


def remove_metric(args: argparse.Namespace) -> None:
    metric = models.Metric.get(models.Metric.name == args.name)

    metric_logs = models.MetricLog.select().where(
        models.MetricLog.metric == metric)

    models.ExportLog.delete().where(
        models.ExportLog.metric_log.in_(metric_logs)).execute()

    models.MetricLog.delete().where(
        models.MetricLog.metric == metric).execute()

    metric.delete_instance()

    print('{name} metric removed'.format(name=metric.name))


def remove_driver(args: argparse.Namespace) -> None:
    driver = models.Driver.get(models.Driver.name == args.name)

    if driver.type == constants.EXPORTER:
        models.ExportLog.delete().where(
            models.ExportLog.driver == driver).execute()

    driver.delete_instance()

    print('{name} driver removed'.format(name=driver.name))


def remove_schedule(args: argparse.Namespace) -> None:
    schedule = models.Schedule.get(models.Schedule.name == args.name)

    models.DriverSchedule.delete().where(
        models.DriverSchedule.schedule == schedule).execute()

    schedule.delete_instance()

    print('{name} schedule removed'.format(name=schedule.name))


def remove_trigger(args: argparse.Namespace) -> None:
    try:
        upper_bound = _bound_to_bool(args.bound)
    except ValueError:
        print(BOUND_INVALID_HELP)
        return

    try:
        powered_on = _power_to_bool(args.power)
    except ValueError:
        print(POWER_INVALID_HELP)
        return

    trigger = models.Trigger.select()\
        .join(models.Driver)\
        .switch(models.Trigger)\
        .join(models.Metric)\
        .where((models.Trigger.power_on == powered_on) &
               (models.Driver.name == args.driver) &
               (models.Trigger.upper_bound == upper_bound) &
               (models.Metric.name == args.metric))\
        .get()

    trigger.delete_instance()

    print('Trigger removed for turning {driver} {power} when {metric} '
          'hits {bound} '.format(
              power=args.power,
              driver=args.driver,
              bound=args.bound,
              metric=args.metric))


def disable_metric(args: argparse.Namespace) -> None:
    metric = models.Metric.get(models.Metric.name == args.name)
    metric.disabled = True
    metric.save()

    print('{name} metric disabled'.format(name=metric.name))


def disable_driver(args: argparse.Namespace) -> None:
    driver = models.Driver.get(models.Driver.name == args.name)
    driver.disabled = True
    driver.save()

    print('{name} driver disabled'.format(name=driver.name))


def disable_schedule(args: argparse.Namespace) -> None:
    schedule = models.Schedule.get(models.Schedule.name == args.name)
    schedule.disabled = True
    schedule.save()

    print('{name} schedule disabled'.format(name=schedule.name))


def disable_trigger(args: argparse.Namespace) -> None:
    try:
        upper_bound = _bound_to_bool(args.bound)
    except ValueError:
        print(BOUND_INVALID_HELP)
        return

    trigger = models.Trigger.select()\
        .join(models.Metric)\
        .where((models.Metric.name == args.metric) &
               (models.Trigger.upper_bound == upper_bound))\
        .get()

    trigger.disabled = True
    trigger.save()

    print('Disabled {bound} trigger for {metric}'.format(
        bound=args.bound, metric=args.metric))


def enable_metric(args: argparse.Namespace) -> None:
    metric = models.Metric.get(models.Metric.name == args.name)
    metric.disabled = False
    metric.save()

    print('{name} metric enabled'.format(name=metric.name))


def enable_driver(args: argparse.Namespace) -> None:
    driver = models.Driver.get(models.Driver.name == args.name)
    driver.disabled = False
    driver.save()

    print('{name} driver enabled'.format(name=driver.name))


def enable_schedule(args: argparse.Namespace) -> None:
    schedule = models.Schedule.get(models.Schedule.name == args.name)
    schedule.disabled = False
    schedule.save()

    print('{name} schedule enabled'.format(name=schedule.name))


def enable_trigger(args: argparse.Namespace) -> None:
    try:
        upper_bound = _bound_to_bool(args.bound)
    except ValueError:
        print(BOUND_INVALID_HELP)
        return

    trigger = models.Trigger.select()\
        .join(models.Metric)\
        .where((models.Metric.name == args.metric) &
               (models.Trigger.upper_bound == upper_bound))\
        .get()

    trigger.disabled = False
    trigger.save()

    print('Enabled {bound} trigger for {metric}'.format(
        bound=args.bound, metric=args.metric))


def add_driver_schedule(args: argparse.Namespace) -> None:
    driver = models.Driver.get(models.Driver.name == args.driver)
    schedule = models.Schedule.get(models.Schedule.name == args.schedule)

    powered_on = None
    if driver.type == constants.POWER:
        if args.power:
            try:
                powered_on = _power_to_bool(args.power)
            except ValueError:
                print(POWER_INVALID_HELP)
                return
        else:
            print(NO_POWER_HELP)
            return

    models.DriverSchedule.create(
        driver=driver,
        schedule=schedule,
        power_on=powered_on
    )

    print('{driver} has been added to {schedule} '
          'which runs: {schedule_text}'.format(
              driver=driver.name,
              schedule=schedule.name,
              schedule_text=schedule.transcribe()
          ))


def remove_driver_schedule(args: argparse.Namespace) -> None:
    driver_schedule = models.DriverSchedule.select()\
            .join(models.Driver)\
            .switch(models.DriverSchedule)\
            .join(models.Schedule)\
            .where((models.Driver.name == args.driver) &
                   (models.Schedule.name == args.schedule))\
            .get()

    driver_schedule.delete_instance()

    print('{driver} has been removed from {schedule}'.format(
        driver=args.driver,
        schedule=args.schedule
    ))


def list_all(args: argparse.Namespace) -> None:
    del args

    print('Metrics: ')
    metrics = models.Metric.select()
    for metric in metrics:
        print('{name} {topic} {type} {manual} {disabled}'.format(
            name=metric.name,
            topic=metric.topic,
            type=metric.type,
            manual='(manual)' if metric.manual else '',
            disabled='(disabled)' if metric.disabled else ''))

    print('\nDrivers: ')
    for dt in constants.drivers:
        drivers = models.Driver.select().where(
            models.Driver.type == dt)

        if drivers:
            print('{type}s:'.format(type=dt))

            for driver in drivers:
                print('{name} {disabled}'.format(
                    name=driver.name,
                    disabled='(disabled)' if driver.disabled else ''))

    print('\nSchedules: ')
    schedules = models.Schedule.select()
    for schedule in schedules:
        print('{name} {schedule} drivers=({drivers}) {disabled}'.format(
            name=schedule.name,
            schedule=schedule.transcribe(),
            drivers=', '.join(['{name} {power}'.format(
                name=driverschedule.driver.name,
                power=_bool_to_power(driverschedule.power_on))
                               for driverschedule
                               in schedule.driver_schedules]),
            disabled='(disabled)' if schedule.disabled else ''
        ))

    print('\nTriggers: ')
    triggers = models.Trigger.select()
    for trigger in triggers:
        print('{metric} reaches {bound}, turn {driver} {power} '
              '{disabled}'.format(
                  metric=trigger.metric.name,
                  bound=_bool_to_bound(trigger.upper_bound),
                  driver=trigger.power_driver.name,
                  power='on' if trigger.power_on else 'off',
                  disabled='(disabled)' if trigger.disabled else ''
              ))


def grow_actions(args: argparse.Namespace) -> None:

    if args.action == 'start':
        try:
            grow.start()
        except grow.GrowAlreadyActiveError:
            print('There is a grow in progress,'
                  'end the current grow to start a new one')
    elif args.action == 'end':
        try:
            grow.end()
        except grow.NoActiveGrowError:
            print('There is no active grow to end')
    else:  # status
        active_grow = grow.get_active()

        if active_grow:
            grow_diff = datetime.utcnow() - active_grow.start
            print('It is currently day {num_days} of the grow'.format(
                num_days=grow_diff.days))

            stage = grow.get_current_stage(active_grow)
            if stage:
                print('Current grow stage is {stage}'.format(stage=stage))
        else:
            print('No active grow')


def manual_actions(args: argparse.Namespace) -> None:

    metric = models.Metric.get_or_none(
        models.Metric.name == args.metric_name)

    if not metric:
        print('{metric_name} metric does not exist'.format(
            metric_name=args.metric_name))
        return

    if metric.manual:
        if args.action == 'on':
            print('{metric_name} already has manual on'.format(
                metric_name=metric.name))
        else:
            metric.manual = False
            metric.save()
            print('{metric_name} set to manual off'.format(
                metric_name=metric.name))
    else:
        if args.action == 'off':
            print('{metric_name} already has manual off'.format(
                metric_name=metric.name))
        else:
            metric.manual = True
            metric.save()
            print('{metric_name} set to manual on'.format(
                metric_name=metric.name))


def read(args: argparse.Namespace) -> None:

    sensors = reflection.load_active_drivers(
        constants.SENSOR,
        args.include,
        args.exclude
    )

    tasks.read(sensors)


def write(args: argparse.Namespace) -> None:

    exporters = reflection.load_active_drivers(
        constants.EXPORTER,
        args.include,
        args.exclude
    )

    tasks.write(exporters)


def power_on(args: argparse.Namespace) -> None:
    power_devices = reflection.load_active_drivers(
        constants.POWER,
        args.drivers
    )

    tasks.power_on(power_devices)


def power_off(args: argparse.Namespace) -> None:
    power_devices = reflection.load_active_drivers(
        constants.POWER,
        args.drivers
    )

    tasks.power_off(power_devices)


class StoreDictKeyPair(argparse.Action):
    """
    Running: `./cli.py --key_pairs 1=2 foo=bar`
    Outputs: `argparse.Namespace(key_pairs={'1': '2', 'foo': 'bar'})`
    Credit: https://stackoverflow.com/a/42355279/1981635
    """
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        self._nargs = nargs
        super(StoreDictKeyPair, self).__init__(option_strings,
                                               dest,
                                               nargs=nargs,
                                               **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        key_pairs = {}

        for kv_str in values:
            key, value = kv_str.split('=')
            key_pairs[key] = value

        setattr(namespace, self.dest, key_pairs)


def create_parser() -> argparse.ArgumentParser:
    main_parser = argparse.ArgumentParser(
        description=LOGO,
        formatter_class=argparse.RawTextHelpFormatter
    )
    main_parser.add_argument('-t', '--test', action='store_true',
                             help='Run in test mode, no data will '
                             'be preserved')
    subparsers = main_parser.add_subparsers()

    new_parser = subparsers.add_parser('new', help='Create a new driver '
                                       'from a template')
    new_parser.set_defaults(func=lambda a: new_parser.print_help())
    new_subparsers = new_parser.add_subparsers()

    new_power_parser = new_subparsers.add_parser(
        'power',
        help='Add a driver for controlling powered devices'
    )
    new_power_parser.add_argument(
        'filename',
        nargs='?',
        help='name of the file to create, the default is {}'.format(
            settings.POWER_DRIVER_TEMPLATE)
    )
    new_power_parser.set_defaults(func=new_power)

    new_sensor_parser = new_subparsers.add_parser(
        'sensor',
        help='Add a sensor for logging sensor readings'
    )
    new_sensor_parser.add_argument(
        'filename',
        nargs='?',
        help='name of the file to create, the default is {}'.format(
            settings.SENSOR_DRIVER_TEMPLATE)
    )
    new_sensor_parser.set_defaults(func=new_sensor)

    new_exporter_parser = new_subparsers.add_parser(
        'exporter',
        help='Add an exporter for exporting logs'
    )
    new_exporter_parser.add_argument(
        'filename',
        nargs='?',
        help='name of the file to create, the default is {}'.format(
            settings.EXPORTER_DRIVER_TEMPLATE)
    )
    new_exporter_parser.set_defaults(func=new_exporter)

    add_parser = subparsers.add_parser('add',
                                       help='Add a metric, driver, schedule, '
                                       'or trigger')
    add_parser.set_defaults(func=lambda a: add_parser.print_help())
    add_subparsers = add_parser.add_subparsers()

    add_metric_parser = add_subparsers.add_parser('metric',
                                                  help='Add a metric')
    add_metric_parser.add_argument('topic', choices=[topic for topic in
                                                     constants.topics],
                                   help='Grouping topic of the metric')
    add_metric_parser.add_argument('type', choices=[metric for metric in
                                                    constants.metrics],
                                   help='The type of metric')
    add_metric_parser.add_argument('name', help='Description of the metric')
    add_metric_parser.add_argument('-m', '--manual', action='store_true',
                                   help='Set the metric as manual')
    add_metric_parser.set_defaults(func=add_metric)

    add_driver_parser = add_subparsers.add_parser('driver',
                                                  help='Add a driver')
    add_driver_parser.add_argument('name', help='Description of the driver')
    add_driver_parser.add_argument('classname',
                                   help='Fully qualified class name (e.g. '
                                   'mymodule.mysubmodule:MyClass)')
    add_driver_parser.add_argument('-c', '--config', dest='config',
                                   action=StoreDictKeyPair,
                                   nargs='+', metavar='KEY=VAL',
                                   help='configuration for the driver')
    add_driver_parser.set_defaults(func=add_driver)

    add_schedule_parser = add_subparsers.add_parser(
        'schedule',
        help='Create a CRON schedule, see https://en.wikipedia.org/wiki/Cron')
    add_schedule_parser.add_argument('name')
    add_schedule_parser.add_argument('minute')
    add_schedule_parser.add_argument('hour')
    add_schedule_parser.add_argument('day_of_month')
    add_schedule_parser.add_argument('month')
    add_schedule_parser.add_argument('day_of_week')
    add_schedule_parser.add_argument('-y', '--yes', action='store_true',
                                     help='Bypass confirmation')
    add_schedule_parser.set_defaults(func=add_schedule)

    add_trigger_parser = add_subparsers.add_parser(
        'trigger',
        help='Setup up a trigger for metric bound')
    add_trigger_parser.add_argument('metric', help='Name of the metric')
    add_trigger_parser.add_argument('bound',
                                    choices=['min', 'max'],
                                    help='Which bound the trigger is for')
    add_trigger_parser.add_argument('driver', type=str)
    add_trigger_parser.add_argument('power', type=str,
                                    choices=['on', 'off'])
    add_trigger_parser.set_defaults(func=add_trigger)

    remove_parser = subparsers.add_parser('remove',
                                          help='Removes a metric, driver '
                                          'schedule, or trigger')
    remove_parser.set_defaults(func=lambda a: remove_parser.print_help())
    remove_subparsers = remove_parser.add_subparsers()

    remove_metric_parser = remove_subparsers.add_parser('metric')
    remove_metric_parser.add_argument('name', type=str, help='Name of the '
                                      'metric to remove. Use the list command '
                                      'to see metric names')
    remove_metric_parser.set_defaults(func=remove_metric)

    remove_driver_parser = remove_subparsers.add_parser('driver')
    remove_driver_parser.add_argument('name', type=str, help='Name of the '
                                      'driver to remove. Use the list command '
                                      'to see driver names')
    remove_driver_parser.set_defaults(func=remove_driver)

    remove_schedule_parser = remove_subparsers.add_parser(
        'schedule',
        help='Remove a schedule')
    remove_schedule_parser.add_argument('name', type=str, help='Name of the '
                                        'schedule to remove. Use the list '
                                        'command to see schedule names')
    remove_schedule_parser.set_defaults(func=remove_schedule)

    remove_trigger_parser = remove_subparsers.add_parser('trigger')
    remove_trigger_parser.add_argument('metric', type=str)
    remove_trigger_parser.add_argument('bound', type=str,
                                       choices=['min', 'max'])
    remove_trigger_parser.add_argument('driver', type=str)
    remove_trigger_parser.add_argument('power', type=str,
                                    choices=['on', 'off'])
    remove_trigger_parser.set_defaults(func=remove_trigger)

    disable_parser = subparsers.add_parser('disable',
                                           help='Disables a metric, driver, '
                                           'schedule, or trigger')
    disable_parser.set_defaults(func=lambda a: disable_parser.print_help())
    disable_subparsers = disable_parser.add_subparsers()

    disable_metric_parser = disable_subparsers.add_parser(
        'metric',
        help='Disables a metric so it will not appear in the Log Entry Service'
    )
    disable_metric_parser.add_argument('name', type=str, help='Name of the '
                                       'metric to disable. Use the list '
                                       'command to see metric names')
    disable_metric_parser.set_defaults(func=disable_metric)

    disable_driver_parser = disable_subparsers.add_parser(
        'driver',
        help='Disables a driver so it will not be executed autonomously'
    )
    disable_driver_parser.add_argument('name', type=str, help='Name of the '
                                       'driver to disable. Use the list '
                                       'command to see driver names')
    disable_driver_parser.set_defaults(func=disable_driver)

    disable_schedule_parser = disable_subparsers.add_parser(
        'schedule',
        help='Disables a schedule so it will not be triggered')
    disable_schedule_parser.add_argument('name', type=str, help='Name of the '
                                         'schedule to disable. Use the list '
                                         'command to see schedule names')
    disable_schedule_parser.set_defaults(func=disable_schedule)

    disable_trigger_parser = disable_subparsers.add_parser('trigger')
    disable_trigger_parser.add_argument('metric', type=str, help='Name of the '
                                        'metric to disable. Use the list '
                                        'command to see schedule names')
    disable_trigger_parser.add_argument('bound', type=str,
                                        choices=['min', 'max'])
    disable_trigger_parser.set_defaults(func=disable_trigger)

    enable_parser = subparsers.add_parser('enable',
                                          help='Enables a metric, driver, '
                                          'schedule, or trigger')
    enable_parser.set_defaults(func=lambda a: enable_parser.print_help())
    enable_subparsers = enable_parser.add_subparsers()

    enable_metric_parser = enable_subparsers.add_parser(
        'metric',
        help='Enables a metric so it will appear in the Log Entry Service, if '
        'set to manual'
    )
    enable_metric_parser.add_argument('name', type=str, help='Name of the '
                                      'metric to enable. Use the list '
                                      'command to see metric names')
    enable_metric_parser.set_defaults(func=enable_metric)

    enable_driver_parser = enable_subparsers.add_parser(
        'driver',
        help='Enables a driver so it will be executed autonomously'
    )
    enable_driver_parser.add_argument('name', type=str, help='Name of the '
                                      'driver to enable. Use the list '
                                      'command to see driver names')
    enable_driver_parser.set_defaults(func=enable_driver)

    enable_schedule_parser = enable_subparsers.add_parser(
        'schedule',
        help='Enables a schedule so it will be triggered'
    )
    enable_schedule_parser.add_argument('name', type=str, help='Name of the '
                                        'schedule to enable. Use the list '
                                        'command to see schedule names')
    enable_schedule_parser.set_defaults(func=enable_schedule)

    enable_trigger_parser = enable_subparsers.add_parser('trigger')
    enable_trigger_parser.add_argument('metric', type=str, help='Name of the '
                                       'metric to enable. Use the list '
                                       'command to see schedule names')
    enable_trigger_parser.add_argument('bound', type=str,
                                       choices=['min', 'max'])
    enable_trigger_parser.set_defaults(func=enable_trigger)

    schedule_parser = subparsers.add_parser('schedule',
                                            help='Attach a driver to a '
                                            'schedule')
    schedule_parser.set_defaults(func=lambda a: schedule_parser.print_help())
    schedule_subparsers = schedule_parser.add_subparsers()

    schedule_driver_help = ('Name of driver. Use the list command to see '
                            'driver names')
    schedule_schedule_help = ('Name of scheduler. Use the list command to see '
                              'schedule names')

    schedule_add_parser = schedule_subparsers.add_parser(
        'add',
        help='Add a driver to a schedule'
    )
    schedule_add_parser.add_argument(
        'driver',
        type=str,
        help=schedule_driver_help
    )
    schedule_add_parser.add_argument(
        'schedule',
        type=str,
        help=schedule_schedule_help
    )
    schedule_add_parser.add_argument(
        'power',
        type=str,
        nargs='?',
        choices=['on', 'off'],
        help='Only used for scheduling power drivers, accepts either on or off'
    )
    schedule_add_parser.set_defaults(func=add_driver_schedule)

    schedule_remove_parser = schedule_subparsers.add_parser(
        'remove',
        help='Remove a driver from a schedule'
    )
    schedule_remove_parser.add_argument(
        'driver',
        type=str,
        help=schedule_driver_help
    )
    schedule_remove_parser.add_argument(
        'schedule',
        type=str,
        help=schedule_schedule_help
    )
    schedule_remove_parser.set_defaults(func=remove_driver_schedule)

    list_parser = subparsers.add_parser('list', help='Lists the metrics with '
                                        'their global IDs, drivers, '
                                        'schedules, and triggers')
    list_parser.set_defaults(func=list_all)

    grow_parser = subparsers.add_parser('grow',
                                        help='Manages the start and end '
                                        'of a grow')
    grow_parser.add_argument('action', choices=['start', 'end', 'status'])
    grow_parser.set_defaults(func=grow_actions)

    manual_parser = subparsers.add_parser('manual',
                                          help=("Change a metric's "
                                                'manual state'))
    manual_parser.add_argument('action', choices=['on', 'off'])
    manual_parser.add_argument('metric_name',
                               help='name of the metric')
    manual_parser.set_defaults(func=manual_actions)

    drivers_include_help = 'Optional list of specific driver to include.'
    drivers_exclude_help = 'Optional list of specific driver to exclude.'

    read_parser = subparsers.add_parser(
        'read',
        help=('Read sensors. If no include of exclude lists are provided, '
              'reads all sensor drivers')
    )
    read_parser.add_argument('-i', '--include', type=str, nargs='+',
                             help=drivers_include_help)
    read_parser.add_argument('-e', '--exclude', type=str, nargs='+',
                             help=drivers_exclude_help)
    read_parser.set_defaults(func=read)

    write_parser = subparsers.add_parser(
        'write',
        help=('Run exporters. If no include of exclude lists are provided, '
              'run all exporters')
    )
    write_parser.add_argument('-i', '--include', type=str, nargs='+',
                              help=drivers_include_help)
    write_parser.add_argument('-e', '--exclude', type=str, nargs='+',
                              help=drivers_exclude_help)
    write_parser.set_defaults(func=write)

    power_parser = subparsers.add_parser('power',
                                         help='Toggle power driver on or off')
    power_parser.set_defaults(func=lambda a: power_parser.print_help())
    power_subparsers = power_parser.add_subparsers()

    power_drivers_help = 'One or more driver names'

    power_on_parser = power_subparsers.add_parser('on',
                                                  help='Powers on drivers')
    power_on_parser.add_argument('drivers', type=str, nargs='+',
                                 help=power_drivers_help)
    power_on_parser.set_defaults(func=power_on)

    power_off_parser = power_subparsers.add_parser('off',
                                                   help='Powers off drivers')
    power_off_parser.add_argument('drivers', type=str, nargs='+',
                                  help=power_drivers_help)
    power_off_parser.set_defaults(func=power_off)

    return main_parser


def create_and_run_parser(
        cli_args: List
) -> (Tuple[argparse.ArgumentParser, argparse.Namespace]):
    parser = create_parser()
    return parser, parser.parse_args(cli_args)


if __name__ == '__main__':
    main()



from typing import List

from gardnr import drivers


def read(sensors: List[drivers.Sensor]) -> None:
    """
    Iterate over every sensor and read from it and create logs and store
    them in database.
    """
    for sensor in sensors:
        # TODO: should be a better way to do this https://goo.gl/xrJdpv
        sensor.read()  # type: ignore



"""
Abstract driver classes as well as common purpose mixins, all custom
drivers must inhereit at least one of these. Drivers hold the code
that interacts with the connected hardware.
"""
from abc import ABCMeta, abstractmethod
from typing import List

from gardnr import constants, logger, models


class Driver(metaclass=ABCMeta):
    """
    Base driver class. Containts features to have "user defined configuration"
    which can be loaded from a JSON string
    """

    def __init__(self, model: models.Driver) -> None:
        """holds user defined configs for drivers and transmitters"""
        self.model = model

        self.setup()

        if self.model.config is not None:
            for key, value in self.model.config.items():
                if hasattr(self, key):
                    logger.warning('{name} is overriding attribute {key} for '
                                   '{qualname}.'.format(
                                       name=self.model.name,
                                       key=key,
                                       qualname=type(self).__qualname__))
                setattr(self, key, value)

    def setup(self) -> None:
        """Do not override __init__, override setup instead."""
        pass

    @property
    @staticmethod
    @abstractmethod
    def type() -> str:
        raise NotImplementedError()


class Power(Driver, metaclass=ABCMeta):
    """supplies power to electrical equiptment"""

    @property
    @staticmethod
    def type() -> str:
        return constants.POWER

    def on(self) -> None:
        """enables power through to connected electrical equiptment"""
        raise NotImplementedError()

    def off(self) -> None:
        """disables power through to connected electrical equiptment"""
        raise NotImplementedError()


class Sensor(Driver, metaclass=ABCMeta):
    """Base sensor specific functions"""

    @property
    @staticmethod
    def type() -> str:
        return constants.SENSOR

    def read(self) -> None:
        """get the readings of the sensor"""
        raise NotImplementedError()


class Exporter(Driver, metaclass=ABCMeta):
    """Base exporter class, all exporter must inherit this."""

    whitelist = None
    blacklist = None

    @property
    @staticmethod
    def type() -> str:
        return constants.EXPORTER

    def export(self, logs: List[models.MetricLog]) -> None:
        """send logs to destination"""
        raise NotImplementedError()



#!/usr/bin/env python

from setuptools import find_packages, setup

setup(
    name='gardnr',
    version='0.5.4',
    author='Jason Biegel',
    url='https://github.com/gardnr/gardnr',
    license='LICENSE',
    description='Monitor and control your grow operation.',
    packages=find_packages(exclude=['docs', 'samples', 'tests']),
    include_package_data=True,
    install_requires=[
        'APScheduler==3.5.3',
        'cron-descriptor==1.2.21',
        'Flask==1.0.2',
        'flask-command==0.0.3',
        'Flask-WTF==0.14.2',
        'gunicorn==19.9.0',
        'grow-recipe==0.8.0',
        'peewee==3.7.1',
        'pytz'
    ],
    entry_points={
        'console_scripts': [
            'gardnr=gardnr.cli:main',
            'gardnr-automata=gardnr.automata:main',
            'gardnr-server=gardnr.server:main'
        ]
    }
)



from gardnr import constants, drivers, logger


class Exporter(drivers.Exporter):
    """
    Uncomment these to filter the types of metrics are logged.
    Either whitelist or blacklist must be used, not both.
    """
    # whitelist = [constants.Metrics.IMAGE]
    # blacklist = [constants.Metrics.IMAGE]

    def setup(self):
        """
        Add configuration here
        """

        # remove the next line and add code
        pass

    def export(self, logs):
        """
        Output the list of logs to an external destination.
        The log object has the following fields available.

        Log:
            id: str
            timestamp: datetime
            longitude: float
            latitude: float
            elevation: float
            value: blob
            metric:
                topic: str
                type: str
                manual: bool
        """
        for log in logs:
            # remove the next line and add code
            pass



import os
import pathlib
from typing import Any, Tuple

from flask import Flask, redirect, render_template, request
from flaskcommand import flask_command
from flask_wtf import FlaskForm, csrf
from wtforms import Field, FileField, FloatField, IntegerField, TextAreaField
from wtforms.validators import Optional as OptionalValidator

from gardnr import constants, metrics, models, settings

app = Flask(__name__)
app_csrf = csrf.CSRFProtect(app)

app.config.update(dict(
    SECRET_KEY='\xc1\x7f;\xbb:\xe6P\xbf0\xa1\x91\xd5X|\xfa\xd4"AZ\xe1/\xe1tt',
    WTF_CSRF_SECRET_KEY='h\xca\xe8\x16>\xae\xd1m\xf9)\xec\xef\x03\x18\xf8\x8d'
))

models.initialize_db()
os.makedirs(settings.UPLOAD_PATH, exist_ok=True)

FIELD_NAME_DELIMITER = ':'

main = flask_command(app)


@app.route('/', methods=('GET', 'POST'))
def manual_logs():
    class MetricForm(FlaskForm):
        """
        Anonymous Form class for constructing
        personalizated forms based on sensors configured
        """
        pass

    # pylint: disable=singleton-comparison
    manual_metrics = models.Metric.select().where(
        (models.Metric.manual == True) &  # noqa: E712
        (models.Metric.disabled == False))

    for metric in manual_metrics:
        metric_field = get_field(metric)

        setattr(MetricForm, metric.name, metric_field)

    # magically populates the form from the request :)
    metric_form = MetricForm()

    # save form
    if metric_form.validate_on_submit():
        for field_name, field_value in metric_form.data.items():
            if field_name == 'csrf_token':
                continue
            if not field_value:
                continue

            metric = models.Metric.get(models.Metric.name == field_name)

            if(metric.type == constants.IMAGE and
               field_value.filename == ''):
                continue

            create_log(metric, field_value)

        return redirect('/')

    return render_template('log_form.html', form=metric_form)


def get_field(metric: models.Metric) -> Field:

    if metric.type == constants.T9E:
        return FloatField(metric.name, validators=[OptionalValidator()])
    elif metric.type == constants.HUMIDITY:
        return IntegerField(metric.name, validators=[OptionalValidator()])
    elif metric.type == constants.IMAGE:
        return FileField(metric.name, validators=[OptionalValidator()])
    elif metric.type == constants.NOTES:
        return TextAreaField(metric.name, validators=[OptionalValidator()])
    elif metric.type == constants.PH:
        return FloatField(metric.name, validators=[OptionalValidator()])
    elif metric.type == constants.EC:
        return FloatField(metric.name, validators=[OptionalValidator()])

    raise ValueError('field type "{field_type}" not found'.format(
        field_type=metric))


def create_log(metric: models.Metric,
               field_value: Any) -> models.MetricLog:

    if metric.type == constants.IMAGE:
        extension = pathlib.Path(field_value.filename).suffix
        image_blob = field_value.read()
        # type: ignore
        return metrics.create_file_log(metric.name, image_blob, extension)

    value = metrics.standardize_metric(metric.type, field_value)

    return metrics.create_metric_log(metric.name, value)
